## -*-Tcl-*-
 # ==========================================================================
 #  WWW Menu - an extension package for Alpha
 # 
 #  FILE: "wwwWindow.tcl"
 #                                    created: 04/30/1997 {11:04:46 am} 
 #                                last update: 05/18/2002 {12:25:08 am} 
 #  Author: Vince Darley
 #  E-mail: <vince@santafe.edu>
 #    mail: 317 Paseo de Peralta, Santa Fe, NM 87501, USA
 #     www: <http://www.santafe.edu/~vince/>
 #     
 #  Description:
 #  
 #  Procedures that support the creation of WWW windows/info.
 #  
 #  See the "wwwVersionHistory.tcl" file for license info, credits, etc.
 # ==========================================================================
 ##

# Make sure that the wwwMode.tcl file has been loaded.
wwwMode.tcl

proc wwwWindow.tcl {} {}

namespace eval WWW {}

# ===========================================================================
# 
#  ----  #
# 
#  WWW Windows  #
# 
# 

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::setWindowParameters" "WWW::setWindowParameters"  --
 # 
 # These are the parameters used for new WWW browser windows.  Whenever the
 # user changes the size of the browser window and either chooses a hyperlink
 # or refreshes/reloads the window, the parameters of the open window are
 # used as the new defaults, so long as the window has been properly passed
 # through 'WWW::renderFile' and does not send its links to a target window.
 # 
 # 'WWW::setWindowParameters' is kind of littered throughout various procs
 # in this package, it would be a good idea to figure out exactly when it
 # needs to be called and remove the rest.
 # 
 # -------------------------------------------------------------------------
 ##

proc WWW::setWindowParameters {{title ""}} {
    
    global mode WWW::WindowParameters win::Active
    
    if {$mode != "WWW"} {
        return
    } elseif {![string length $title]} {
	set title [win::Current]
    } elseif {![lcontains win::Active $title]} {
	return
    }
    set g [getGeometry $title]
    set WWW::WindowParameters($title) $g
    if {![string length [WWW::linksToTarget $title]]} {
	set WWW::WindowParameters(Default) $g
    }
}

proc WWW::getWindowParameters {{title ""}} {
    
    global WWW::WindowParameters
    
    if {![string length $title]} {set title [win::Current]}
    if {[info exists WWW::WindowParameters($title)]} {
	set WWW::WindowParameters($title)
    } else {
	set WWW::WindowParameters(Default)
    }
}

## 
 # -------------------------------------------------------------------------
 # 
 # "WWW::setFillColumn"  --
 # 
 # Determine the fill column length.
 # 
 # If the WWW mode pref 'Auto Adjust Width' is not set, adjusting the WWW
 # mode pref 'fillColumn' will change the length of the text line strings in
 # the window.  Otherwise, since we know that the window was created with
 # 'WWW::createNewWindow', the font settings are known.  The formula used to
 # determine the fill column variable in this case probably could be fine
 # tuned, but works reasonably well for now.
 #  
 # -------------------------------------------------------------------------
 ##

proc WWW::setFillColumn {{title ""}} {
    
    global WWWmodeVars alpha::platform fontSize

    if {![string length $title]} {set title [win::Current]}
    set w [lindex [getGeometry $title] 2] ; # l t w h
    if {!$WWWmodeVars(autoAdjustWidth)} {
	set fc $WWWmodeVars(fillColumn)
    } elseif {[info tclversion] < 8.0} {
	# Alpha7
	set fc [expr {($w / $fontSize) * 1.35}]
    } elseif {[set alpha::platform] == "alpha"} {
	# Alpha8/X -- needs more work, needs more support !
	set fc [expr {($w / $WWWmodeVars(wwwFontSize)) * 1.35}]
    } elseif {[set alpha::platform] == "tk"} {
	# Alphatk
	set f  $WWWmodeVars(wwwFont)
	set fs $WWWmodeVars(wwwFontSize)
	set fm [font measure [list $f $fs] "_"]
	set fc [expr {($w / $fm) - 4}]
    } else {
	# ???
	set fc $WWWmodeVars(fillColumn)
    }
    expr {int($fc)}
}

proc WWW::createNewWindow {title} {

    global WWWmodeVars mode win::Active WWW::WindowParameters \
      WWW::LinkTargetWindow WWW::LinkWindowTarget
    
    # Should we kill the current window? ...
    set window  [win::Current]
    # Is the current window part of a frameset target?
    foreach item [array names WWW::LinkWindowTarget] {
	if {[set WWW::LinkWindowTarget($item)] == $window} {
	    set targetTo $item
	    break
	} 
    }
    if {[string length [set target [WWW::linksToTarget $window]]]} {
	if {[info exists WWW::LinkWindowTarget($target)]} {
	    set targetWindow [set WWW::LinkWindowTarget($target)]
	} else {
	    set targetWindow ""
	}
        if {[lcontains win::Active $targetWindow]} {
	    WWW::setWindowParameters $targetWindow
	    if {!$WWWmodeVars(linksOpenNewWindow) && $window != "* WWW History *"} {
		bringToFront $targetWindow
		killWindow
	    }
	} 
    } elseif {$mode == "WWW"} {
	if {!$WWWmodeVars(linksOpenNewWindow) && $window != "* WWW History *"} {
	    WWW::setWindowParameters
	    killWindow
	}
    }
    set g [WWW::getWindowParameters $title]
    if {[info tclversion] < 8.0} {
	eval [list new -n $title -m "WWW" -shell 1 -g] $g
    } else {
        set f  $WWWmodeVars(wwwFont)
	set fs $WWWmodeVars(wwwFontSize)
	eval [list new -n $title -m "WWW" -shell 1 -font $f -fontsize $fs -g] $g
    }
    set window [win::Current]
    set WWW::LinkWindowTarget($target) $window
    if {[info exists targetTo]} {
        set WWW::LinkWindowTarget($targetTo) $window
    } 
    set window
}

proc WWW::linksToTarget {title} {
    
    global win::Modes WWW::LinkTargetWindow

    if {[catch {set win::Modes($title)} winMode]} {
        return ""
    } elseif {$winMode != "WWW"} {
	return ""
    } elseif {![info exists WWW::LinkTargetWindow($title)]} {
	return ""
    } else {
	set WWW::LinkTargetWindow($title)
    }
}

proc WWW::reload {title {fromCache 0}} {

    global WWW::FileSource WWW::UrlSource
    
    if {[info exists WWW::FileSource($title)]} {
	if {$fromCache || ![temp::isIn WWW-tmp [set WWW::FileSource($title)]]} {
	    set script [list WWW::renderFile [set WWW::FileSource($title)]]
	} else {
	    file delete [set WWW::FileSource($title)]
	    unset WWW::FileSource($title)
	}
    }
    if {![info exists script]} {
	if {[info exists WWW::UrlSource($title)]} {
	    if {!$fromCache} {
		if {[file isfile $title]} {
		    file delete $title
		}
	    }
	    set script [list WWW::renderUrl [set WWW::UrlSource($title)]]
	} else {
	    status::errorMsg "Couldn't find the source of '$title'."
	}
    }
    
    # Remember where we are ...
    set pos [pos::fromRowCol [win::getInfo $title currline] 0]
    # ... then refresh the file from the source.
    killWindow
    eval $script
    goto $pos ; insertToTop
}

# Used, e.g., by Wiki mode.
proc WWW::webPageHasChanged {url {newContents ""}} {

    global WWW::UrlSource WWW::FileSource

    foreach name [array names WWW::UrlSource] {
	if {[set WWW::UrlSource($name)] == $url} {
	    set localfile $name
	    break
	}
    }
    if {![info exists localfile]} { return }
    
    if {[catch {bringToFront $localfile}]} {
	foreach name [array names WWW::FileSource] {
	    if {[set WWW::FileSource($name)] == $localfile} {
		set localwindow $name
		break
	    }
	}
	if {![info exists localwindow]} { return }
    } else {
	set localwindow $localfile
    }
    
    catch {bringToFront $localwindow}
    if {[info exists WWW::FileSource($localwindow)]} {
	if {[string length $newContents]} {
	    file::writeAll [set WWW::FileSource($localwindow)] $newContents 1
	    WWW::reload $localwindow 1
	} else {
	    WWW::reload $localwindow
	}
    }
    
}

# ===========================================================================
# 
# .